/*
 * deviced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * @file	auto-brightness-sensorhub.h
 * @brief	auto-brightness-sensorhub header file
 */
#ifndef __AUTO_BRT_SENSORHUB_H__
#define __AUTO_BRT_SENSORHUB_H__

#define SPECIFIC_MODE_OFF	0
#define LBM_LEVEL		110
#define HBM_LEVEL		120

#include "display/display.h"

int prepare_level_handler(void);
void exit_level_handler(void);

/* Control brightness considering various situations overlap each other, such as
 * HBM on/off, LBM on/off, Hold on/off, LowDim on/off, DimStay, Set
 *
 * [State and Flag]
 * Brightness is managed by 4 flag
 *   L  H Ho Ld
 *   0  0  0  0
 *   (L: LBM / H: HBM / Ho: Hold / Ld: Lowbattery Dim)
 *
 * State is combination of those 4 flags, for example
 *   0010 -> Only hold brightness is enabled
 *   1011 -> LBM, Hold, Lowbattery dim enabled
 *
 * There are state pairs that cannot be both enabled.
 *   LBM and HBM cannot be both enabled.
 *     - On eabling one, always disables the other one.
 *   HBM and Hold cannot be both enabled.
 *     - An request that arrives first takes priority over request arrives later.
 *       First arriving one is enabled and the later one is ignored.
 *
 * And there is a special state related to brightness, abnormal battery health.
 * This is not managed by using flag because it only makes brightness dim temporarily,
 * and all controlling brightness works the way it was regardless of whether
 * the abnormal battery health is happened. The only thing to do is deciding the brightness
 * to restore when battery health returned to normal, default or hold brightness. For example,
 *   Default(70) -> Abnormal(Dim) -> Normal : restore to default brightness 70
 *   Default(70) -> Hold(100) -> Abnormal(Dim) -> Normal : restore to hold brightness 100
 *
 * [Request]
 * There are 9 requests, 4 are enabling request, 4 are disabling request, and 1 is set request.
 *   Enabling request : LBM on, HBM on, Hold, LowbatteryDim on
 *   Disabling request: LBM off, HBM off, Release, LowbatteryDim off
 *   Set request: set default brightness and changes brightness to default. Set by user.
 *
 * [Policy]
 * Define action(flag setting and brightness change) by request and state.
 *
 *  Request LBM on
 *   - State 0000, 0001: enable LBM. change brightness to default(0000) or dim(0001)
 *   - State 0010, 0011: do not enable LBM as someone holds brightness
 *   - Set L flag to the above states
 *
 *  Request HBM on
 *   - State 0000, 0001: Only these state can be enablied with HBM.
 *                       Change brightness to MAX and enable HBM.
 *   - Set H flag to the above states
 *
 *  Request Hold
 *   - State 0000, 0001: change brightness to hold brightness
 *   - State 0010, 0011, 1010, 1011: As Hold flag has already been set, just change
 *                       brightness to request one
 *   - State 1000, 1001: If LBM has been enabled and nobody holds brightness, then
 *                       disable LBM first and apply Hold brightness
 *     Set Ho flag to the above states
 *
 *  Request Lowbattery Dim on
 *   - State 0000, 1000: change brightness to dim
 *   - State 0010, 0100, 1010: do nothing as someone request HBM/Hold request. At these
 *                       states, stay that brightness, not to change to dim brightness
 *   - Set Ld flag to the above states
 *
 *   Request LBM off
 *    - State 1000, 1001: disable LBM, change brightness to default(1000) or dim(1001)
 *    - State 1010, 1011: do nothing as these state, LBM have already been disabled as
 *                        someone had requested HBM/Hold. Stay that brightness
 *    - Unset L flag to the above states
 *
 *   Request HBM off
 *    - State 0100, 0101: disable HBM, change brightness to default(0010) or dim(0011)
 *    - Unset H flag to the above states
 *
 *   Request Release
 *    - State 0010, 0011: change brightness to default(0010) or dim(0011)
 *    - State 1010, 1011: enable LBM, change brightness to default(1010) or dim(1011)
 *    - Unset Ho flag to the above states
 *
 *   Request Lowbattery Dim off
 *    - State 0001, 1001: change brightness to default
 *    - State 0011, 0101, 1011: do nothing as someone request HBM/Hold request
 *    - Unset Ld flag to the above states
 *
 *   Request brightness set
 *    - For all possible states, Set default brightness, Update brightness vconf
 *    - For states that HBM disabled, change brightness to default brightness.
 *    - Unset Ho, Ld flag to the states that H flag isn't set.
 */
int auto_brightness_control(enum brightness_request_e request, int set_brightness);

/* Restore brightness on recovering battery health to normal */
int auto_brightness_restore(void);

#endif
