/*
 * deviced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <fcntl.h>

#include "auto-brightness-sensorhub.h"
#include "weaks.h"
#include "display-info.h"
#include "display/util.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "core/common.h"
#include "core/device-notifier.h"
#include "core/config-parser.h"

#define BOARD_CONF_FILE "/etc/deviced/display.conf"

#define ON		"on"
#define OFF		"off"

#define SIGNAL_HBM_ON	"HBMOn"
#define SIGNAL_HBM_OFF	"HBMOff"

#define HBM_LEVEL	120

#define LCD_PATH    "sys/class/lcd/"
#define HBM_PATH    "/hbm"
#define DEVICE_PATH "/device"
#define PATH_BUFFER_MAX    512

static guint timer;
static struct timespec offtime;
static char *hbm_path;

static void broadcast_hbm_state(char *state)
{
	dbus_handle_emit_dbus_signal(NULL,
			DEVICED_PATH_DISPLAY,
			DEVICED_INTERFACE_DISPLAY,
			state,
			NULL);
}

static void hbm_set_offtime(int timeout)
{
	struct timespec now;

	if (timeout <= 0) {
		offtime.tv_sec = 0;
		return;
	}

	clock_gettime(CLOCK_REALTIME, &now);
	offtime.tv_sec = now.tv_sec + timeout;
}

static gboolean hbm_off_cb(void *data)
{
	timer = 0;

	if (pm_cur_state != S_NORMAL) {
		_D("Hbm timeout, but it's not display normal.");
		return G_SOURCE_REMOVE;
	}
	hbm_set_offtime(0);

	auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);

	return G_SOURCE_REMOVE;
}

static void hbm_start_timer(int timeout)
{
	if (timer) {
		g_source_remove(timer);
		timer = 0;
	}
	if (!timer) {
		timer = g_timeout_add_seconds(timeout, hbm_off_cb, NULL);
	}
}

static void hbm_end_timer(void)
{
	if (timer) {
		g_source_remove(timer);
		timer = 0;
	}
}

int hbm_get_state(void)
{
	char state[5];
	int ret, hbm;

	if (!hbm_path)
		return -ENODEV;

	ret = sys_get_str(hbm_path, state, sizeof(state));
	if (ret < 0)
		return ret;

	if (!strncmp(state, ON, strlen(ON)))
		hbm = true;
	else if (!strncmp(state, OFF, strlen(OFF)))
		hbm = false;
	else
		hbm = -EINVAL;

	return hbm;
}

int hbm_set_state(int hbm)
{
	int ret;
	if (!hbm_path)
		return -ENODEV;

	if (hbm)
		broadcast_hbm_state(SIGNAL_HBM_ON);
	else
		broadcast_hbm_state(SIGNAL_HBM_OFF);

	ret = sys_set_str(hbm_path, (hbm ? ON : OFF));
	if (ret < 0)
		_E("Failed to HBM %s.", hbm ? "ON" : " OFF");

	return ret;
}

static int hbm_set_state_with_timeout(int hbm, int timeout)
{
	int ret;

	if (hbm && (timeout <= 0))
		return -EINVAL;

	if (hbm)
		ret = auto_brightness_control(BR_HBM_ON, BR_IMPLICIT);
	else
		ret = auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);

	if (ret < 0)
		return ret;

	_D("timeout is %d", timeout);

	if (hbm) {
		/*
		 * hbm is turned off after timeout.
		 */
		hbm_set_offtime(timeout);
		hbm_start_timer(timeout);
	} else {
		hbm_set_offtime(0);
		hbm_end_timer();
		broadcast_hbm_state(SIGNAL_HBM_OFF);
	}

	return 0;
}

static void hbm_check_timeout(void)
{
	struct timespec now;

	if (timer) {
		g_source_remove(timer);
		timer = 0;
	}

	if (offtime.tv_sec == 0) {
		_E("It's invalid state. HBM is turned off.");
		auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		return;
	}

	clock_gettime(CLOCK_REALTIME, &now);
	_D("now %ld, offtime %ld", now.tv_sec, offtime.tv_sec);

	/* check it's timeout */
	if (now.tv_sec >= offtime.tv_sec) {
		hbm_set_offtime(0);
		auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
	} else {
		_D("HBM state is restored.");
		auto_brightness_control(BR_HBM_ON, BR_IMPLICIT);
		hbm_start_timer(offtime.tv_sec - now.tv_sec);
	}
}

static void hbm_get_level(GVariant *var, void *user_data, GError *err)
{
	int level;

	if (!var) {
		_D("Invalid parameter.");
		return;
	}

	if (!dh_get_param_from_var(var, "(i)", &level)) {
		_E("Failed to get message, %s.", g_variant_get_type_string(var));
		goto out;
	}

	if (level == HBM_LEVEL) {
		_I("Lux was high already, HBM enable");
		auto_brightness_control(BR_HBM_ON, BR_IMPLICIT);
	}
out:
	g_variant_unref(var);
}

static gboolean hbm_check_handler(gpointer data)
{
	int ret;

	ret = dbus_handle_method_async_with_reply(COORD_BUS_NAME,
			COORD_PATH_AUTOBRIGHTNESS,
			COORD_INTERFACE_AUTOBRIGHTNESS,
			"GetLevel",
			NULL,
			NULL,
			hbm_get_level,
			-1,
			NULL);

	if (ret < 0)
		_D("Failed to call coord.autobrightness.GetLevel method, %d", ret);

	return G_SOURCE_REMOVE;
}

static int display_state_changed(void *data)
{
	int state;
	int ret;

	state = DATA_VALUE_INT(data);

	if (get_outdoor_setting)
		if (get_outdoor_setting())
			return 0;

	switch (state) {
	case S_NORMAL:
		/*
		 * outdoor-enhance-mode not supported
		 *  : check & restore hbm always.
		 * outdoor-enhance-mode supported
		 *  : check & restore hbm when old state is dim only.
		 */
		if (!get_outdoor_setting || pm_old_state == S_LCDDIM)
			hbm_check_timeout();

		(void) g_timeout_add(300, hbm_check_handler, NULL);
		break;
	case S_LCDDIM:
	case S_LCDOFF:
	case S_SLEEP:
		ret = auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		if (ret < 0)
			_E("Failed to off hbm.");
		hbm_end_timer();
		break;
	}

	return 0;
}

static GVariant *dbus_gethbm(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return g_variant_new("(i)", hbm_get_state());
}

static GVariant *dbus_sethbm_timeout(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int hbm, timeout, ret;

	g_variant_get(param, "(ii)", &hbm, &timeout);

	if (timeout < 0) {
		_E("Timeout can not be setting(%d).", timeout);
		ret = -EINVAL;
		goto out;
	}

	ret = hbm_set_state_with_timeout(hbm, timeout);

	if (ret < 0)
		_E("Failed to set HBM (ret=%d, hbm=%d, time=%d).", ret, hbm, timeout);
	else
		_I("Set hbm (ret=%d, hbm=%d, time=%d).", ret, hbm, timeout);

out:
	return g_variant_new("(i)", ret);
}

static const dbus_method_s dbus_methods[] = {
	{ "GetHBM",          NULL,   "i", dbus_gethbm },
	{ "SetHBMTimeout",   "ii",   "i", dbus_sethbm_timeout },
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_DISPLAY,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int hbm_func(unsigned int cmd, void *arg)
{
	int ret = 0;
	struct hbmsetstate *hss;

	switch (cmd) {
	case HBM_GET_STATE:
		ret = hbm_get_state();
		break;
	case HBM_SET_STATE:
		if ((int *)arg)
			ret = auto_brightness_control(BR_HBM_ON, BR_IMPLICIT);
		else
			ret = auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		break;
	case HBM_TURN_ON:
		ret = auto_brightness_control(BR_HBM_ON, BR_IMPLICIT);
		break;
	case HBM_TURN_OFF:
	case HBM_TURN_OFF_STATE:
		ret = auto_brightness_control(BR_HBM_OFF, BR_IMPLICIT);
		break;
	case HBM_SET_TIMEOUT_STATE:
		hss = (struct hbmsetstate *)arg;
		ret = hbm_set_state_with_timeout(hss->hbm, hss->timeout);
		break;
	default:
		ret = -EINVAL;
		break;
	}

	return ret;
}

static char *check_and_copy_path(char *path)
{
	int fd;
	char *buf = NULL;

	fd = open(path, O_RDONLY);
	if (fd >= 0) {
		buf = strndup(path, strlen(path));
		close(fd);
	} else {
		_E("Failed to open HBM node.");
	}

	return buf;
}

static char *find_hbm_node()
{
	DIR *d;
	struct dirent *dir;
	char buf[PATH_BUFFER_MAX];
	char *result = NULL;

	d = opendir(LCD_PATH);
	if (!d)
		return NULL;

	while ((dir = readdir(d))) {
		if (dir->d_name[0] == '.')
			continue;

		snprintf(buf, sizeof(buf), "%s%s%s", LCD_PATH,
			dir->d_name, HBM_PATH);

		result = check_and_copy_path(buf);
		if (result)
			break;

		snprintf(buf, sizeof(buf), "%s%s%s%s", LCD_PATH,
			dir->d_name, DEVICE_PATH, HBM_PATH);

		result = check_and_copy_path(buf);
		if (result)
			break;
	}
	closedir(d);

	return result;
}

static void hbm_init(void *data)
{
	int ret;

	hbm_path = find_hbm_node();

	if (!hbm_path) {
		_E("Failed to find HBM node.");
		return;
	} else {
		_I("HBM node: %s.", hbm_path);
	}

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_DISPLAY, &dbus_interface);
	if (ret < 0)
		_E("Failed to register dbus object.");

	/* register notifier */
	register_notifier(DEVICE_NOTIFIER_LCD, display_state_changed);
}

static void hbm_exit(void *data)
{
	/* unregister notifier */
	unregister_notifier(DEVICE_NOTIFIER_LCD, display_state_changed);

	/*
	 * set default brightness
	 * if display logic is stopped in hbm state.
	 */
	if (hbm_get_state() == true) {
		hbm_set_offtime(0);
		_I("set brightness to default value!");
	}
}

static const struct display_ops display_hbm_ops = {
	.name     = "hbm",
	.init     = hbm_init,
	.exit     = hbm_exit,
	.func     = hbm_func,
};

DISPLAY_OPS_REGISTER(&display_hbm_ops)

