/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <vconf.h>
#include <device-node.h>
#include <hw/led.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"
#include "core/devices.h"
#include "torch.h"

#define LED_MAX_BRIGHTNESS      100
#define GET_BRIGHTNESS(x)       (((x) >> 24) & 0xFF)

#define SIGNAL_FLASH_STATE "ChangeFlashState"

static struct led_device *led_dev;
static struct led_state led_state = {
	.type = LED_TYPE_MANUAL,
	.color = 0x0,
	.duty_on = 0,
	.duty_off = 0,
};

static void flash_state_broadcast(int val)
{
	int ret;

	ret = dbus_handle_emit_dbus_signal(NULL,
						DEVICED_PATH_LED,
						DEVICED_INTERFACE_LED,
						SIGNAL_FLASH_STATE,
						g_variant_new("(i)", val));
	if (ret < 0)
		_E("Failed to send dbus signal(%s)", SIGNAL_FLASH_STATE);
}

static GVariant *dbus_get_brightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, alpha;

	if (!led_dev) {
		_E("There is no led device.");
		ret = -ENOENT;
		goto error;
	}

	alpha = GET_BRIGHTNESS(led_state.color);
	ret = alpha * 100.f / 255;
	if (alpha != 0 && alpha != 0xFF)
		ret += 1;
	_D("Get brightness. color=%#x: %d", led_state.color, ret);

error:
	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_max_brightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = LED_MAX_BRIGHTNESS;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_set_brightness(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int val, enable, ret;
	struct led_state tmp = {0,};

	g_variant_get(param, "(ii)", &val, &enable);

	if (!led_dev) {
		_E("There is no led device.");
		ret = -ENOENT;
		goto error;
	}

	tmp.color = (((int)(val * 255.f) / LED_MAX_BRIGHTNESS) & 0xFF) << 24;
	_D("Set brightness. val=%d color=%#x", val, tmp.color);

	ret = led_dev->set_state(&tmp);
	if (ret < 0)
		goto error;

	memcpy(&led_state, &tmp, sizeof(led_state));

	/* flash status broadcast */
	flash_state_broadcast(val);

	/* if enable is ON, noti will be show or hide */
	if (enable) {
		if (val)
			ongoing_show();
		else
			ongoing_clear();
	}

error:
	return g_variant_new("(i)", ret);
}

/**
	* GetBrightnessForCamera is for camera library.
	* Currently they do not have a daemon for camera,
	* but they need to get camera brightness value without led priv.
	* It's a temporary solution on Tizen 2.4 and will be removed asap.
	*/
static const dbus_method_s dbus_methods[] = {
	{ "GetBrightnessForCamera", NULL,   "i", dbus_get_brightness },
	{ "GetBrightness",    NULL,   "i", dbus_get_brightness },
	{ "GetMaxBrightness", NULL,   "i", dbus_get_max_brightness },
	{ "SetBrightness",    "ii",   "i", dbus_set_brightness },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_LED,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};


static int led_service_load(void)
{
	struct hw_info *info;
	int r;

	r = hw_get_info(LED_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (r < 0) {
		_E("Failed to load led shared library: %d", r);
		return -ENOENT;
	}

	if (!info->open) {
		_E("Fail to open camera led device: open(NULL)");
		return -EPERM;
	}

	r = info->open(info, LED_ID_CAMERA_BACK,
			(struct hw_common **)&led_dev);
	if (r < 0) {
		_E("Failed to get camera led device: %d", r);
		return -EPERM;
	}

	_D("Camera led device structure load success.");
	return 0;
}

static int led_service_free(void)
{
	struct hw_info *info;

	if (!led_dev)
		return -ENOENT;

	info = led_dev->common.info;

	assert(info);

	info->close((struct hw_common *)led_dev);

	return 0;
}

static int torch_probe(void *data)
{
	/* load led device */
	return led_service_load();
}

static void torch_init(void *data)
{
	int ret;

	/* init dbus interface */
	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_LED, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);
}

static void torch_exit(void *data)
{
	/* free led device */
	led_service_free();
}

static const struct device_ops torchled_device_ops = {
	DECLARE_NAME_LEN("torchled"),
	.probe    = torch_probe,
	.init     = torch_init,
	.exit     = torch_exit,
};

DEVICE_OPS_REGISTER(&torchled_device_ops)
