/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define FUSE_USE_VERSION 26

#define _GNU_SOURCE
#include <semaphore.h>
#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unzip.h>
#include <pthread.h>
#include <signal.h>
#include <gio/gio.h>
#include <glib.h>
#include <sys/stat.h>
#include <assert.h>
#include <attr/xattr.h>
#include <linux/limits.h>
#include <tzplatform_config.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"
#include "core/devices.h"
#include "core/list.h"
#include "core/device-notifier.h"
#include "core/common.h"
#include "tzip-utility.h"

#define PACKAGE_UNPACK_PATH		"/opt/usr/share/package-unpacked"

enum tzip_state_t {
	TZIP_RUNNING_STATE = 0,
	TZIP_STOP_STATE,
	TZIP_ERROR_STATE,
};

static pthread_t thread;
static pthread_t mount_thread;
static pthread_attr_t attr;
static struct fuse *fuse_handle;
static struct fuse_chan *channel;
static GAsyncQueue *async_queue;
static gint stop_or_error = TZIP_RUNNING_STATE;

static int thread_ready;
static pthread_cond_t thread_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t thread_cond_mutex = PTHREAD_MUTEX_INITIALIZER;
static volatile bool tzip_frozen = false;

static int tzip_is_mounted(const char *mount_point);
static int tzip_unmount_zipfs(const char *mount_point);
static int tzip_mount_zipfs(const char *src_file, const char *mount_point, const char *smack);

static int check_path_available(char *mountpath)
{
	const char *path;

	if (!mountpath)
		return 0;

	path = tzplatform_getenv(TZ_SYS_HOME);
	if (path && !strncmp(mountpath, path, strlen(path)))
		return 1;

	path = tzplatform_getenv(TZ_SYS_RW_APP);
	if (path && !strncmp(mountpath, path, strlen(path)))
		return 1;

	if (!strncmp(mountpath, PACKAGE_UNPACK_PATH, strlen(PACKAGE_UNPACK_PATH)))
		return 1;

	return 0;
}

static int tzip_getattr(const char *path, struct stat *stbuf)
{
	int res = 0;

	if (!path || !stbuf) {
		_E("Invalid Arguments path=%p stbuf=%p.", path, stbuf);
		return -EINVAL;
	}

	_D("Get ATTR path(%s).", path);

	tzip_lock();
	res = get_path_prop(path, stbuf);
	tzip_unlock();

	_D("Exit: %d\n", res);

	return res;
}

static int tzip_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
			 off_t offset, struct fuse_file_info *fi)
{
	struct tzip_dir_info *flist;
	GHashTableIter iter;
	gpointer key, value;
	char *name;
	int ret;

	if (!path || !buf) {
		_E("Invalid Arguments path=%p buf=%p.", path, buf);
		return -EINVAL;
	}

	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);

	tzip_lock();
	flist = get_dir_files(path);
	if (flist == NULL) {
		_E("No Files in Dir(%s).", path);
		ret = 0;
		goto out_unlock;
	}

	g_hash_table_iter_init(&iter, flist->file_hash);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		name = (char *)key;
		filler(buf, name, NULL, 0);
	}

	ret = 0;

out_unlock:
	tzip_unlock();

	return ret;
}

static int tzip_open(const char *path, struct fuse_file_info *fi)
{
	int ret;
	char *file = NULL;
	char *zippath = NULL;
	int dir_status;
	struct tzip_mount_entry *entry = NULL;
	struct tzip_handle *handle = NULL;
	unz_file_info *file_info = NULL;
	unz_global_info global_info;
	unzFile *zipfile = NULL;
	int len;
	int file_len;

	if (!path || !fi) {
		_E("Invalid Arguments  path=%p fi=%p.", path, fi);
		return -EINVAL;
	}
	_D("Open path=%s", path);

	tzip_lock();
	entry = find_mount_entry(path, &dir_status);
	if (!entry) {
		_E("Failed to find mount entry.");
		ret = -ENOENT;
		goto out_unlock;
	}

	zippath = strdup(entry->zip_path);
	if (!zippath) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out_unlock;
	}

	len = strlen(entry->path);
	file_len = strlen(path) - len + 1;
	file = (char *)malloc(file_len);
	if (!file) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out_unlock;
	}
	strncpy(file, path + len + 1, file_len);
	tzip_unlock();

	zipfile = unzOpen(zippath);
	if (!zipfile) {
		_E("Failed to unzOpen.");
		ret = -ENOENT;
		goto out;
	}

	/* Get info about the zip file */
	if (unzGetGlobalInfo(zipfile, &global_info) != UNZ_OK) {
		_E("Failed to unzGetGlobalInfo.");
		ret = -EINVAL;
		goto out;
	}

	if (unzLocateFile(zipfile, file, CASE_SENSITIVE) != UNZ_OK) {
		_E("Failed to unzLocateFile: file(%s) is not Found", file);
		ret = -ENOENT;
		goto out;
	}

	file_info = (unz_file_info *)malloc(sizeof(unz_file_info));
	if (!file_info) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out;
	}

	ret = unzGetCurrentFileInfo(zipfile, file_info, (char *)file,
			strlen(file), NULL, 0, NULL, 0);
	if (ret != UNZ_OK) {
		_E("Failed to unzGetCurrentFileInfo.");
		ret = -EINVAL;
		goto out;
	}

	ret = unzOpenCurrentFile(zipfile);
	if (ret != UNZ_OK) {
		_E("Failed to unzOpenCurrentFile.");
		ret = -EINVAL;
		goto out;
	}
	handle = (struct tzip_handle *)malloc(sizeof(struct tzip_handle));
	if (!handle) {
		unzCloseCurrentFile(zipfile);
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out;
	}
	handle->file_info = file_info;
	handle->zipfile = zipfile;
	handle->offset = 0;
	handle->path = zippath;
	handle->file = file;
	handle->pbuf = NULL;
	handle->from = 0;
	handle->to = 0;
	if (sem_init(&handle->lock, 0, 1) == -1)
		_E("Failed to sem_init.");

	fi->fh = (uint64_t)(unsigned long)handle;

	return 0;

out_unlock:
	tzip_unlock();
out:
	free(zippath);
	free(file);
	free(file_info);
	if (zipfile)
		unzClose(zipfile);
	return ret;
}

static int tzip_read(const char *path, char *buf, size_t size, off_t offset,
				struct fuse_file_info *fi)
{
	struct tzip_handle *handle;
	int ret;

	if (!path || !buf) {
		_E("Invalid Arguments path=%p buf=%p.", path, buf);
		return -EINVAL;
	}
	if (!fi || fi->fh == 0) {
		_E("Invalid Zip Handle.");
		return -EINVAL;
	}

	handle = (struct tzip_handle *)(unsigned long)(fi->fh);

	_D("Read Path=%s size=%zd offset=%jd.", path, size, offset);
	sem_wait(&handle->lock);
	ret = read_zipfile(handle, buf, size, offset);
	sem_post(&handle->lock);
	_D("Read ret=%d.", ret);
	return ret;
}

static int tzip_release(const char *path, struct fuse_file_info *fi)
{
	struct tzip_handle *handle;

	if (!path) {
		_E("Invalid arguments path=%p.", path);
		return -EINVAL;
	}
	if (!fi || fi->fh == 0) {
		_E("Invalid zip handle.");
		return -EINVAL;
	}

	handle = (struct tzip_handle *)(unsigned long)(fi->fh);

	if (sem_destroy(&handle->lock) == -1)
		_E("Failed to sem_destroy.");

	unzCloseCurrentFile(handle->zipfile);
	unzClose(handle->zipfile);
	free(handle->file);
	free(handle->path);
	free(handle->file_info);
	if (handle->pbuf)
		free(handle->pbuf);
	free(handle);

	return 0;
}

static struct fuse_operations tzip_oper = {
	.getattr = tzip_getattr,
	.open = tzip_open,
	.read = tzip_read,
	.release = tzip_release,
	.readdir = tzip_readdir,
};

static void *tzip_thread(void *arg)
{
	int ret;
	FILE *fp;
	char *file_path;
	char *mount_dir;
	char *mount_info = NULL;
	size_t len;
	ssize_t read;
	char *saveptr = NULL;
	char *argv[4] = {"deviced", "-o", "allow_other,kernel_cache,auto_cache", NULL};
	struct fuse_args args = FUSE_ARGS_INIT(3, argv);

	ret = mkdir(TZIP_ROOT_PATH, 0755);
	if (ret < 0 && errno != EEXIST) {
		_E("Failed to make dir(%s).", TZIP_ROOT_PATH);
		goto out;
	}

	tzip_lock();
	channel = fuse_mount(TZIP_ROOT_PATH, &args);
	if (!channel) {
		_E("Trying to mount after cleaning up previous instance '%p'.", TZIP_ROOT_PATH);
		fuse_unmount(TZIP_ROOT_PATH, NULL);

		channel = fuse_mount(TZIP_ROOT_PATH, NULL);
		if (!channel) {
			_E("Failed to fuse mount(%p).", TZIP_ROOT_PATH);
			goto out_unlock;
		}
	}

	fuse_handle = fuse_new(channel, &args, &tzip_oper, sizeof(tzip_oper), NULL);
	if (!fuse_handle) {
		_E("Failed to fuse_new.");
		goto out_unlock;
	}

	/* initialize string hash table */
	if (!hashmap_init()) {
		_E("Failed to hashmap_init.");
		goto out_unlock;
	}

	fp = fopen(TZIP_INFO_FILE, "r");
	if (fp) {
		while ((read = getline(&mount_info, &len, fp)) != -1) {
			if (mount_info) {
				mount_info[strcspn(mount_info, "\n")] = '\0';
				file_path = (char *)strtok_r(mount_info, ":", &saveptr);
				mount_dir = (char *)strtok_r(NULL, ":", &saveptr);

				if (file_path != NULL && mount_dir != NULL)
					tzip_remount_zipfs(file_path, mount_dir);
				free(mount_info);
				mount_info = NULL;
			}
		}
		fclose(fp);
	}
	tzip_unlock();

	/* notify to tzip_mount_thread */
	pthread_mutex_lock(&thread_cond_mutex);
	thread_ready = 1;
	pthread_cond_signal(&thread_cond);
	pthread_mutex_unlock(&thread_cond_mutex);

	if (fuse_loop_mt(fuse_handle)) {
		_E("Failed to fuse_loop_mt.");
		g_atomic_int_set(&stop_or_error, TZIP_ERROR_STATE);
		return NULL; /* already send pthread_cond_signal() */
	}

	tzip_lock();
	if (fuse_handle) {
		fuse_destroy(fuse_handle);
		fuse_handle = NULL;
	}
	tzip_unlock();

	return NULL;

out_unlock:
	tzip_unlock();
out:
	g_atomic_int_set(&stop_or_error, TZIP_ERROR_STATE);

	/* notify to tzip_mount_thread */
	pthread_mutex_lock(&thread_cond_mutex);
	thread_ready = 1;
	pthread_cond_signal(&thread_cond);
	pthread_mutex_unlock(&thread_cond_mutex);

	return NULL;
}

static void free_tzip_msg_data(struct tzip_msg_data *msg)
{
	if (msg) {
		if (msg->zippath)
			free(msg->zippath);

		if (msg->mountpath)
			free(msg->mountpath);

		if (msg->smack)
			free(msg->smack);

		free(msg);
	}
}

#define TZIP_MSGTYPE_STR(type)    \
	type == 'm' ? "Mount" :   \
	type == 'u' ? "Unmount" : \
	type == 'i' ? "IsMount" : "Unknown Command"

static void *tzip_mount_thread(void *arg)
{
	struct tzip_msg_data *msgdata;
	int ret = 0;

	/* wait for tzip_thread to be ready*/
	pthread_mutex_lock(&thread_cond_mutex);
	if (!thread_ready)
		pthread_cond_wait(&thread_cond, &thread_cond_mutex);
	assert(thread_ready); /* Never happend */
	pthread_mutex_unlock(&thread_cond_mutex);

	while (1) {
		/* g_async_queue_pop() is a blocking call, it should never return NULL */
		msgdata = (struct tzip_msg_data *)g_async_queue_pop(async_queue);
		assert(msgdata);

		switch (msgdata->type) {
		case 'm':
			ret = tzip_mount_zipfs(msgdata->zippath, msgdata->mountpath, msgdata->smack);
			break;

		case 'u':
			ret = tzip_unmount_zipfs(msgdata->mountpath);
			break;

		case 'i':
			ret = tzip_is_mounted(msgdata->mountpath);
			break;

		default:
			ret = -EINVAL;
			break;
		}

		if (ret < 0)
			_E("Failed to %s operation. path=%s: %d", TZIP_MSGTYPE_STR(msgdata->type), msgdata->mountpath, ret);
		else
			_D("Successful %s operation. path=%s.", TZIP_MSGTYPE_STR(msgdata->type), msgdata->mountpath);

		if (msgdata->invocation) {
			/* g_object_unref(msgdata->invocation) is called within this function */
			g_dbus_method_invocation_return_value(msgdata->invocation, g_variant_new("(i)", ret));
		}

		free_tzip_msg_data(msgdata);
	}

	/* dead code, added just to satisfy compiler */
	return NULL;
}

static void tzip_server_init(void)
{
	static int once = 0;

	/* You do not need pthread_once because it runs only in the main loop. */
	if (once)
		return;
	once = 1;

	_D("tzip_server_init");

	/* for tzip_mount_thread */
	async_queue = g_async_queue_new();
	if (!async_queue) {
		_E("Failed to create async_queue.");
		goto error;
	}

	if (pthread_attr_init(&attr) != 0) {
		_E("Failed to pthread_attr_init.");
		goto error;
	}

	if (pthread_create(&thread, &attr, &tzip_thread, NULL)) {
		_E("Failed to pthread_create.");
		goto error;
	}

	if (pthread_create(&mount_thread, &attr, &tzip_mount_thread, NULL)) {
		_E("Failed to pthread_create.");
		goto error;
	}

	return;

error:
	g_atomic_int_set(&stop_or_error, TZIP_ERROR_STATE);
}

static void tzip_server_exit(void)
{
	if (!fuse_handle || !channel)
		return;

	tzip_lock();
	fuse_exit(fuse_handle);
	fuse_unmount(TZIP_ROOT_PATH, channel);
	channel = NULL;
	tzip_unlock();

	if (pthread_join(thread, NULL) != 0)
		_E("Failed to pthread_join.");
}

static int tzip_check_mount_point(const char *mount_point)
{
	int ret;

	if (unlink(mount_point) == 0)
		return 0;

	ret = errno;
	if (ret == ENOENT)
		return 0;

	_E("Failed to remove previous symlink file: %d", ret);
	return ret;
}

static int tzip_mount_zipfs(const char *src_file, const char *mount_point, const char *smack)
{
	int ret = 0;
	char *tzip_path = NULL;

	tzip_lock();
	ret = add_mount_entry(src_file, mount_point);
	if (ret) {
		_E("Failed to add_mount_entry '%s'.", mount_point);
		goto out;
	}

	ret = tzip_check_mount_point(mount_point);
	if (ret < 0) {
		_E("Failed to mount via tzip: %d", ret);
		remove_mount_entry(mount_point);
		goto out;
	}

	if (sizeof(TZIP_ROOT_PATH) + strlen(mount_point) > PATH_MAX) {
		_E("path length exceed %d. mount_point(%zu).", PATH_MAX, strlen(mount_point));
		remove_mount_entry(mount_point);
		ret = -ENOMEM;
		goto out;
	}

	ret = asprintf(&tzip_path, "%s%s", TZIP_ROOT_PATH, mount_point);
	if (ret < 0) {
		_E("Failed to asprintf.");
		remove_mount_entry(mount_point);
		ret = -ENOMEM;
		goto out;
	}

	_D("Creating sym link: '%s' and '%s'.", tzip_path, mount_point);
	ret = symlink(tzip_path, mount_point);
	if (ret) {
		ret = -errno;
		_E("Failed to symlink.");
		remove_mount_entry(mount_point);
		goto out;
	}

	ret = tzip_store_mount_info(src_file, mount_point);
	if (ret != 0)
		_E("Failed to store_mount_info '%s'.", mount_point);

	ret = lsetxattr(mount_point, "security.SMACK64", smack, strlen(smack), 0);
	if (ret < 0) {
		ret = -errno;
		_E("Failed to set smack label using lsetxattr().");
		remove_mount_entry(mount_point);
		ret = unlink(mount_point);
		if (ret) {
			ret = -errno;
			_E("Failed to unlink '%s'.", mount_point);
		}
	}

out:
	tzip_unlock();
	free(tzip_path);
	_D("Exit: %d", ret);
	return ret;
}

static int tzip_unmount_zipfs(const char *mount_point)
{
	int ret = 0;
	char *tzip_path = NULL;

	tzip_lock();
	ret = remove_mount_entry(mount_point);
	if (ret) {
		_E("Failed to remove_mount_entry '%s'.", mount_point);
		goto out_unlock;
	}

	if (sizeof(TZIP_ROOT_PATH) + strlen(mount_point) > PATH_MAX) {
		_E("path length exceed %d. mount_point(%zu).", PATH_MAX, strlen(mount_point));
		ret = -ENOMEM;
		goto out_unlock;
	}

	ret = asprintf(&tzip_path, "%s%s", TZIP_ROOT_PATH, mount_point);
	if (ret < 0) {
		_E("Failed to asprintf.");
		ret = -ENOMEM;
		goto out_unlock;
	}

	_D("Deleting sym link: '%s' and '%s'.", tzip_path, mount_point);
	ret = unlink(mount_point);
	if (ret) {
		ret = -errno;
		_E("Failed to unlink.");
	}

	if (tzip_clear_mount_info())
		_E("Failed to clear_mount_info '%s'.", mount_point);

out_unlock:
	tzip_unlock();
	free(tzip_path);
	_D("Exit: %d", ret);
	return ret;
}

static int tzip_is_mounted(const char *mount_point)
{
	int ret = 0;
	char *tzip_path = NULL;
	struct stat sb;

	if (tzip_frozen) {
		_E("Tzip is currently frozen.");
		return -EAGAIN;
	}

	tzip_lock();
	if (!get_hashmap()) {
		_E("Tzip init is not done yet.");
		ret = -EOWNERDEAD;
		goto out_unlock;
	}

	if (!get_mount_entry(mount_point)) {
		_E("mount_path(%s) does not exists.", mount_point);
		ret = 0;
		goto out_unlock;
	}
	tzip_unlock();

	if (sizeof(TZIP_ROOT_PATH) + strlen(mount_point) > PATH_MAX) {
		_E("path length exceed %d. mount_point(%zu).", PATH_MAX, strlen(mount_point));
		return -ENOMEM;
	}

	ret = asprintf(&tzip_path, "%s%s", TZIP_ROOT_PATH, mount_point);
	if (ret < 0) {
		_E("Failed to asprintf.");
		return -ENOMEM;
	}

	if (stat(tzip_path, &sb) == -1 || stat(mount_point, &sb) == -1)
		ret = 0;
	else
		ret = 1;

	free(tzip_path);
	return ret;

out_unlock:
	tzip_unlock();
	return ret;
}

static GVariant *dbus_request_mount_tzip_internal(GDBusConnection *conn, const gchar *sender, GVariant *param,
						  GDBusMethodInvocation *invocation, bool sync)
{
	int ret;
	pid_t pid;
	char *smack;
	char *zippath;
	char *mountpath;
	GVariant *result;
	struct tzip_msg_data *msgdata = NULL;

	tzip_server_init();
	if (g_atomic_int_get(&stop_or_error) != TZIP_RUNNING_STATE) {
		_E("Tzip is not running state.");
		ret = -EOWNERDEAD;
		goto out;
	}

	g_variant_get(param, "(&s&s&s)", &mountpath, &zippath, &smack);

	if (!mountpath || !zippath || !smack) {
		_E("Failed to g_variant_get.");
		ret = -ENOMEM;
		goto out;
	}

	if (!*mountpath || !*zippath || !*smack) {
		_E("Invalid input.");
		ret = -EINVAL;
		goto out;
	}

	if (!check_path_available(mountpath)) {
		_E("TZIP mount path(%s) is invalid.", mountpath);
		ret = -EAGAIN;
		goto out;
	}

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (is_app(pid) != 0) {
		_E("PID(%d) is not privileged to use tzip.", pid);
		ret = -EPERM;
		goto out;
	}

	if (tzip_frozen) {
		_E("Tzip is currently frozen.");
		ret = -EAGAIN;
		goto out;
	}

	msgdata = calloc(1, sizeof(struct tzip_msg_data));
	if (!msgdata) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out;
	}

	msgdata->type = 'm';
	msgdata->smack = strdup(smack);
	msgdata->zippath = strdup(zippath);
	msgdata->mountpath = strdup(mountpath);
	if (!msgdata->smack || !msgdata->zippath || !msgdata->mountpath) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out;
	}

	ret = 0;

	if (sync) {
		result = NULL;
		msgdata->invocation = invocation;
	} else
		result = g_variant_new("(i)", ret);

	g_async_queue_push(async_queue, (gpointer)msgdata);

out:
	if (ret < 0) {
		free_tzip_msg_data(msgdata);
		result = g_variant_new("(i)", ret);
	}

	return result;
}

static GVariant *dbus_request_unmount_tzip_internal(GDBusConnection *conn, const gchar *sender, GVariant *param,
						    GDBusMethodInvocation *invocation, bool sync)
{
	int ret;
	pid_t pid;
	char *mountpath;
	GVariant *result;
	struct tzip_msg_data *msgdata = NULL;

	tzip_server_init();
	if (g_atomic_int_get(&stop_or_error) != TZIP_RUNNING_STATE) {
		_E("Tzip is not running state.");
		ret = -EOWNERDEAD;
		goto out;
	}

	g_variant_get(param, "(&s)", &mountpath);

	if (!mountpath) {
		_E("Failed to g_variant_get.");
		ret = -ENOMEM;
		goto out;
	}

	if (!*mountpath) {
		_E("Invalid input.");
		ret = -EINVAL;
		goto out;
	}

	if (!check_path_available(mountpath)) {
		_E("TZIP mount path(%s) is invalid.", mountpath);
		ret = -EAGAIN;
		goto out;
	}

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (is_app(pid) != 0) {
		_E("PID(%d) is not privileged to use tzip.", pid);
		ret = -EPERM;
		goto out;
	}

	if (tzip_frozen) {
		_E("Tzip is currently frozen.");
		ret = -EAGAIN;
		goto out;
	}

	msgdata = calloc(1, sizeof(struct tzip_msg_data));
	if (!msgdata) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out;
	}

	msgdata->type = 'u';
	msgdata->mountpath = strdup(mountpath);
	if (!msgdata->mountpath) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out;
	}

	ret = 0;

	if (sync) {
		result = NULL;
		msgdata->invocation = invocation;
	} else
		result = g_variant_new("(i)", ret);

	g_async_queue_push(async_queue, (gpointer)msgdata);

out:
	if (ret < 0) {
		free_tzip_msg_data(msgdata);
		result = g_variant_new("(i)", ret);
	}

	return result;
}

static GVariant *dbus_request_mount_tzip(GDBusConnection *conn,
					const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
					GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return dbus_request_mount_tzip_internal(conn, sender, param, invocation, false);
}

static GVariant *dbus_request_unmount_tzip(GDBusConnection *conn,
					const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
					GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return dbus_request_unmount_tzip_internal(conn, sender, param, invocation, false);
}

static GVariant *dbus_request_mount_sync_tzip(GDBusConnection *conn,
					      const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
					      GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return dbus_request_mount_tzip_internal(conn, sender, param, invocation, true);
}

static GVariant *dbus_request_unmount_sync_tzip(GDBusConnection *conn,
						const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
						GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return dbus_request_unmount_tzip_internal(conn, sender, param, invocation, true);
}

static GVariant *dbus_request_ismounted_tzip(GDBusConnection *conn,
					const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
					GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	char *mountpath;
	struct tzip_msg_data *msgdata = NULL;

	tzip_server_init();
	if (g_atomic_int_get(&stop_or_error) != TZIP_RUNNING_STATE) {
		_E("Tzip is not running state.");
		ret = -EOWNERDEAD;
		goto out;
	}

	g_variant_get(param, "(&s)", &mountpath);

	if (!mountpath) {
		_E("Failed to g_variant_get.");
		ret = -ENOMEM;
		goto out;
	}

	if (!*mountpath) {
		_E("Invalid input.");
		ret = -EINVAL;
		goto out;
	}

	msgdata = calloc(1, sizeof(struct tzip_msg_data));
	if (!msgdata) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out;
	}

	msgdata->type = 'i';
	msgdata->mountpath = strdup(mountpath);
	msgdata->invocation = invocation;
	if (!msgdata->mountpath) {
		_E("Failed to malloc.");
		ret = -ENOMEM;
		goto out;
	}

	g_async_queue_push(async_queue, (gpointer)msgdata);
	ret = 0;
out:
	if (ret < 0) {
		free_tzip_msg_data(msgdata);
		return g_variant_new("(i)", ret);
	}

	return NULL;
}

static const dbus_method_s dbus_methods[] = {
	{ "Mount", "sss", "i", dbus_request_mount_tzip },
	{ "Unmount", "s", "i", dbus_request_unmount_tzip },
	{ "MountSync", "sss", "i", dbus_request_mount_sync_tzip },
	{ "UnmountSync", "s", "i", dbus_request_unmount_sync_tzip },
	{ "IsMounted", "s", "i", dbus_request_ismounted_tzip },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_TZIP,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int booting_done(void *data)
{
	/* To reduce the latency of the first Tzip operation, you can apply this optimization.
	tzip_server_init(); */
	return 0;
}

static int tzip_poweroff(void *data)
{
	_I("TZIP poweroff.");
	tzip_server_exit();
	return 0;
}

#define FREEZER_TZIP_PATH	"/sys/fs/cgroup/freezer/tzip"
static void tzip_init(void *data)
{
	int ret;
	_D("tzip_init");

	tzip_lock_init();

	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	register_notifier(DEVICE_NOTIFIER_POWEROFF, tzip_poweroff);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_TZIP, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);

	ret = mkdir(FREEZER_TZIP_PATH, 0755);
	if (ret < 0 && errno != EEXIST)
		_E("Failed to create freezer tzip.");
}

static void tzip_exit(void *data)
{
	_D("tzip_exit");
	tzip_server_exit();
	unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	unregister_notifier(DEVICE_NOTIFIER_POWEROFF, tzip_poweroff);

	tzip_lock_deinit();
}

#define TZIP_USER_PATH	"/run/tzip/opt/usr"
static void tzip_suspend(void)
{
	tzip_frozen = true;
	suspend_path("tzip", TZIP_USER_PATH, 5);

	/* FIXME
	Between tzip_suspend() and actual system suspend (autosleep),
	additional Tzip user processes might appear.
	We still need to fix this race condition problem as well.
	*/
}

static void tzip_resume(void)
{
	tzip_frozen = false;
	resume_path("tzip", TZIP_USER_PATH);
}

static const struct device_ops tzip_device_ops = {
	DECLARE_NAME_LEN("tzip"),
	.init    = tzip_init,
	.exit    = tzip_exit,
	.suspend = tzip_suspend,
	.resume  = tzip_resume
};

DEVICE_OPS_REGISTER(&tzip_device_ops)

